#========================================================================================
# package      DE_EPAGES::Database::API::Config
# state        public
# description  This module contains access function to read Database.d/$Store.conf.
#========================================================================================
package DE_EPAGES::Database::API::Config;
use base qw (Exporter);
use strict;

our @EXPORT_OK  = qw(
    StoreConnections
    StoreType
    StoreTypes
    StoreTypesWithCartridge
    BUStoreTypes
    ShopStoreTypes
    GetBaseStoreType
    BusinessUnits
    GetBusinessUnits
    SetBusinessUnits
    DatabaseConfig
    DatabaseConfigFileName
    DatabaseConnectionData
    DatabaseBackupData
    DatabaseRedirect
    GetRootConnections
);

use DES::EDE3CBC qw (des_3ede_cbc_encrypt);
use DE_EPAGES::Core::API::Error qw (Error ExistsError);
use DE_EPAGES::Core::API::File qw (GetCachedFile RunOnDirectory RunOnGlob GetFileContent WriteFile);
use DE_EPAGES::Core::API::IniConfig;
use DE_EPAGES::Core::API::Log qw (GetLog LogDebug);
use DE_EPAGES::Core::API::PerlTools qw( Hex2Str trim unique UsePackage );
use DE_EPAGES::Core::API::SystemCommand qw( SystemCommand );
use DE_EPAGES::Core::API::TestParameter qw( TestString );
use DE_EPAGES::Core::API::Warning qw ( Warning );
use DE_EPAGES::XML::API::XML::CartridgeInstallerImportHandler qw(GetCartridges GetStoreTypeCartridges);



sub ErrorDefinition {
    return {
        'VoidConfigKey' => 'Key "#Key=#Value" with void value in section "#Section" in file "#FileName"',
        'MissingConfigKey' => 'Missing key "#Key" in section "#Section" in file "#FileName"',
        'MissingConfigSection' => 'Missing section "#Section" in file "#FileName"',
        'NoStoreType' => 'No store type found for "#StoreName" in file "#FileName"',
        'FileReadFailed' => 'Can\'t read file #FileName',
        'BackupSectionAndKeyDefined' => 'Ini file contains both Backup section and Backup key: #FileName',
    };
}

#========================================================================================
# function     StoreConnections
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aStoreConnections = StoreConnections();
#----------------------------------------------------------------------------------------
# description  get all connections that have config file Database.d/$Store.conf
#----------------------------------------------------------------------------------------
# return       $aStoreConnections | list of basenames from Database.d | ref.array
#========================================================================================
sub StoreConnections () {

    my @aStoreConnections;
    if (-d "$ENV{EPAGES_CONFIG}/Database.d") {
        RunOnDirectory( "$ENV{EPAGES_CONFIG}/Database.d", sub {
            foreach my $Conf (glob('*.conf')) {
                if (not -r $Conf) {
                    Warning('FileReadFailed', { 'FileName' => $Conf });
                    next;
                }
                $Conf =~ s/\.conf$//;
                $Conf =~ s,.*/,,;
                push @aStoreConnections, $Conf;
            }
        });
    }
    return \@aStoreConnections;
}

#========================================================================================
# function     StoreType
# state        public
#----------------------------------------------------------------------------------------
# syntax       $StoreType = StoreType( $Store );
#----------------------------------------------------------------------------------------
# description  Returns the store type of a store
#----------------------------------------------------------------------------------------
# input        $Store | logical database name | string
# return       $StoreType | store type of $Store | string
#========================================================================================
sub StoreType ($) {
    my ($Store) = @_;
    TestString('Store', $Store);

    my $hSection = DatabaseConnectionData($Store, { incomplete => 1 });
    Error('NoStoreType', {
        'StoreName' => $Store,
        'FileName' => DatabaseConfigFileName($Store),
    }) unless defined $hSection->{StoreType};

    return $hSection->{StoreType};
}

#========================================================================================
# function     StoreTypes
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aStoreTypes = StoreTypes();
#----------------------------------------------------------------------------------------
# description  Returns a list of all store types that have at least 1 store
#----------------------------------------------------------------------------------------
# return       $aStoreTypes | list of list store types | ref.array
#========================================================================================
sub StoreTypes () {

    return [
        unique
        grep { defined $_ and $_ ne 'Backup'}
        map { DatabaseConnectionData($_, { incomplete => 1 })->{'StoreType'} }
        @{StoreConnections()}
    ];
}

#========================================================================================
# function     StoreTypesWithCartridge
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aStoreTypeList = StoreTypesWithCartridge($CartridgeName);
# example      $aStoreTypeList = StoreTypesWithCartridge('DE_EPAGES::Shop');
#----------------------------------------------------------------------------------------
# description  Returns a list of store types that have cartridge $CartridgeName and
#               at least 1 store using the store type or config file in StoreTypes.d/.
#----------------------------------------------------------------------------------------
# input        $CartridgeName  | name of a cartridge | string
# return       $aStoreTypeList | list of store types | ref.array
#========================================================================================
sub StoreTypesWithCartridge ($) {
    my ($CartridgeName) = @_;

    my $hCartridgesXml = GetCartridges();
    my @StoreTypes = @{StoreTypes()};

    return [
        grep { $CartridgeName ~~ $hCartridgesXml->{$_} and $_ ~~ @StoreTypes }
        keys %$hCartridgesXml
    ];
}

#========================================================================================
# function     ShopStoreTypes
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aStoreTypeList = ShopStoreTypes();
#----------------------------------------------------------------------------------------
# description  Returns a list of store types that have cartridge "DE_EPAGES::Shop".
#----------------------------------------------------------------------------------------
# return       $aStoreTypeList | list of store types | ref.array
#========================================================================================
sub ShopStoreTypes {
    return StoreTypesWithCartridge('DE_EPAGES::Shop');
}

#========================================================================================
# function     GetBaseStoreType
# state        public
#----------------------------------------------------------------------------------------
# syntax       ($BaseStoreType, $AdditionalCartridges) = GetBaseStoreType($StoreType, $BaseStoreTypeConfigDir);
#----------------------------------------------------------------------------------------
# description  Returns the $BaseStoreType (one of: Site Store Monitor Master MultiStore)
#               for $StoreType depending on $BaseStoreTypeConfigDir/$BaseStoreType.conf:
#               1. returns nothing if $StoreType has stores with dialect
#               2. use $BaseStoreType.conf if has same cartridges as $StoreType
#               3. use $BaseStoreType.conf if has all cartridges as $StoreType (min. superset)
#               4. use $BaseStoreType.conf if has all cartridges as $StoreType which have
#                  no Data/Public/lib/ (other cartridges are $AdditionalCartridges)
#               5. otherwise returns nothing
#----------------------------------------------------------------------------------------
# input        $StoreType | name of store types | string
# input        $BaseStoreTypeConfigDir | name of dir containing $BaseStoreType.conf | string
# input        $IgnoreDialects | do not return if dialects found | boolean
# return       $BaseStoreType | base store type | string
# return       $AdditionalCartridges | $StoreType cartridges which are not part of base
#                  store type but without Data/Public/lib/ | ref.array
#========================================================================================
sub GetBaseStoreType ($$;$) {
    my ($StoreType, $BaseStoreTypeConfigDir, $IgnoreDialects) = @_;

    # find dialects in $ENV{EPAGES_CONFIG}/StoreTypes.d/VERSION/$StoreType.conf
    my $CartridgePackage = 'DE_EPAGES::Core::API::Cartridge';
    UsePackage($CartridgePackage);
    my $Version = $CartridgePackage->new()->version;
    my $ConfigFile = "$ENV{EPAGES_CONFIG}/StoreTypes.d/$Version/$StoreType.conf";
    if (not $IgnoreDialects and -f $ConfigFile) {
        my $IniConfig = DE_EPAGES::Core::API::IniConfig->new(
            FileName => $ConfigFile, CaseSensitive => 1);
        my @Dialects = grep {/-/} keys %{$IniConfig->section('Languages')};
        # return nothing if $StoreType has dialects
        return if scalar @Dialects;
    }

    # UsePackage because this package and the used package require each other
    UsePackage('DE_EPAGES::Installer::API::LanguagePack');
    # eval { GetUsedLanguages } due to $BusinessUnits w/o running databases
    my $BusinessUnits = GetBusinessUnits({conf => 1});
    my ($hStoreTypeLanguages) =
        eval { DE_EPAGES::Installer::API::LanguagePack::GetUsedLanguages(
                   $BusinessUnits, {Ignore => [qw(de en)]})
    };

    if (not $IgnoreDialects and not ExistsError() and
        defined $hStoreTypeLanguages->{$StoreType}) {
        my @Dialects = grep {/-/} keys %{$hStoreTypeLanguages->{$StoreType}};
        # return nothing if $StoreType has stores with dialect
        return if scalar @Dialects;
    }

    # create %StoreTypeCartridges with key == $StoreType
    # and value == sorted list of cartridges for $StoreType
    my %StoreTypeCartridges;
    RunOnGlob($BaseStoreTypeConfigDir, '*.conf', sub {
        my $IniConfig = DE_EPAGES::Core::API::IniConfig->new(
            FileName => "$BaseStoreTypeConfigDir/$_",
            CaseSensitive => 1,
            );
        s,\.conf$,,;
        $StoreTypeCartridges{$_} = [ sort keys %{$IniConfig->section('Cartridges')} ];
    });

    # @SortedStoreTypes is sorted by number of cartridges per store type
    my @SortedStoreTypes = sort { scalar @{$StoreTypeCartridges{$a}}
                                  <=> scalar @{$StoreTypeCartridges{$a}}
    } keys %StoreTypeCartridges;

    # first come: Store Site Monitor Master MultiStore
    my @SortedBaseStoreTypes;
    foreach (qw(Site Store Monitor Master MultiStore)) {
        push @SortedBaseStoreTypes, $_ if $_ ~~ @SortedStoreTypes;
    }
    # then come other
    my @SortedOtherStoreTypes =
        grep { !/^(Site|Store|Monitor|Master|MultiStore)$/ } @SortedStoreTypes;
    @SortedStoreTypes = (@SortedBaseStoreTypes, @SortedOtherStoreTypes);

    my @StoreTypeCartridges = sort @{GetStoreTypeCartridges($StoreType)};

    # first try: $BaseStoreType.conf has same cartridges as $StoreType

    foreach my $BaseStoreType (($StoreType, @SortedStoreTypes)) {
        next unless defined $StoreTypeCartridges{$BaseStoreType};
        return ($BaseStoreType)
            if @StoreTypeCartridges ~~ @{$StoreTypeCartridges{$BaseStoreType}};
    }

    # second try: $BaseStoreType.conf has all cartridges as $StoreType

    foreach my $BaseStoreType (($StoreType, @SortedStoreTypes)) {
        next unless defined $StoreTypeCartridges{$BaseStoreType};
        my $Success = 1;
        foreach my $Cartridge (@StoreTypeCartridges) {
            if (not $Cartridge ~~ @{$StoreTypeCartridges{$BaseStoreType}}) {
                $Success = 0;
                last;
            }
        }
        return ($BaseStoreType) if $Success;
    }

    # third try: $BaseStoreType.conf has all cartridges as $StoreType
    #            which have no Data/Public/lib

    foreach my $BaseStoreType (($StoreType, @SortedStoreTypes)) {
        next unless defined $StoreTypeCartridges{$BaseStoreType};
        my $Success = 1;
        my $AdditionalCartridges = [];
        foreach my $Cartridge (@StoreTypeCartridges) {
            if (not $Cartridge ~~ @{$StoreTypeCartridges{$BaseStoreType}}) {
                my $CartridgeDir = $Cartridge;
                $CartridgeDir =~ s,::,/,g;
                my $DataPublicLib = "$ENV{EPAGES_CARTRIDGES}/$CartridgeDir/Data/Public/lib";
                if (-d $DataPublicLib) {
                    $Success = 0;
                    last;
                } else {
                    push @$AdditionalCartridges, $Cartridge;
                }
            }
        }
        return ($BaseStoreType, $AdditionalCartridges) if $Success;
    }

    # otherwise: no $BaseStoreType found

    return;
}

#========================================================================================
# function     BUStoreTypes
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aStoreTypes = BUStoreTypes();
#----------------------------------------------------------------------------------------
# description  Returns a list of all store types that have a business unit
#----------------------------------------------------------------------------------------
# return       $aStoreTypes | list of store types | ref.array
#========================================================================================
sub BUStoreTypes () {
    my $BU = BusinessUnits();
    return [ unique map { $BU->{$_}{Connection}{StoreType} } keys %$BU ];
}

#========================================================================================
# function     GetBusinessUnits
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aBusinessUnits = GetBusinessUnits($hFlags);
# syntax       $aBusinessUnits = GetBusinessUnits({env => 1, conf => 1});
#----------------------------------------------------------------------------------------
# description  Returns a list of all stores that are business units.
#----------------------------------------------------------------------------------------
# input        $hFlags | define the information source for the list
#                   <li>env -  get list from environment variable EPAGES_BUSINESS_UNITS
#                   <li>conf - get list from config files Database.d/*.conf
#                   <li>(both/empty) - from EPAGES_BUSINESS_UNITS (if set) or from conf
#               </ul> | ref.hash
# return       $aBusinessUnits | list of business unit names | ref.array.string
#========================================================================================
sub GetBusinessUnits {
    my ($hFlags) = @_;

    my $ReturnEnv = $ENV{EPAGES_BUSINESS_UNITS} =~ /\S/;
    $ReturnEnv = 1 if $hFlags->{env} and not $hFlags->{conf};
    $ReturnEnv = 0 if $hFlags->{conf} and not $hFlags->{env};

    return [ sort split /\s+/, $ENV{EPAGES_BUSINESS_UNITS} ] if $ReturnEnv;
    return [ sort keys %{BusinessUnits()} ];
}

#========================================================================================
# function     SetBusinessUnits
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Env = SetBusinessUnits($aBusinessUnits);
# example      $Env = SetBusinessUnits(GetBusinessUnits({conf => 1}));
#----------------------------------------------------------------------------------------
# description  sets the the environment variable EPAGES_BUSINESS_UNITS to the space join
#               of $aBusinessUnits in:
#               - Linux: /etc/default/epages6
#               - Windows: registry
#               returns the new value of EPAGES_BUSINESS_UNITS
#----------------------------------------------------------------------------------------
# input        $aBusinessUnits | list of business unit names | ref.array.string
# return       $Env | new value of EPAGES_BUSINESS_UNITS | string
#========================================================================================
sub SetBusinessUnits {
    my ($aBusinessUnits) = @_;
    $ENV{EPAGES_BUSINESS_UNITS} = join(' ', @{$aBusinessUnits}); ## no critic qw(RequireLocalizedPunctuationVars)

    if ($^O eq 'MSWin32') {
        SystemCommand(['setx', 'EPAGES_BUSINESS_UNITS', $ENV{EPAGES_BUSINESS_UNITS}, '-m']);
    } else {
        my $rContent = GetFileContent("/etc/default/$ENV{EPAGES_PROJECT}");
        $$rContent =~ s/\n\s*(EPAGES_BUSINESS_UNITS=).*/\n$1"$ENV{EPAGES_BUSINESS_UNITS}"/g;
        $$rContent .= 'export EPAGES_BUSINESS_UNITS'
            unless $$rContent =~ /\n\s*export .*EPAGES_BUSINESS_UNITS/;
        WriteFile("/etc/default/$ENV{EPAGES_PROJECT}", $rContent);
    }
    return $ENV{EPAGES_BUSINESS_UNITS};
}

#========================================================================================
# function     BusinessUnits
#----------------------------------------------------------------------------------------
# syntax       $StoreConf = BusinessUnits();
#----------------------------------------------------------------------------------------
# description  Gets a hash of store information from Database.d/*.conf for all stores
#               with parameters StoreType, Database, datasource
#               and datasource =~ /mysql/ and Database !~ /mysql|epagesj|aspooldb|systemdb/
#----------------------------------------------------------------------------------------
# return       $StoreConf | hash of store information | ref.hash
#========================================================================================
sub BusinessUnits {

    my ($ConfFile, %StoreConf);

    foreach my $Store (@{StoreConnections()}) {
        my %DatabaseConf;
        $ConfFile = DatabaseConfigFileName($Store);
        next unless -s $ConfFile;
        tie %DatabaseConf, 'Config::IniFiles', (-file => $ConfFile) or
            Error('FileReadFailed', { 'FileName' => $ConfFile });

        my %CartridgeConf;
        $ConfFile = "$ENV{EPAGES_CONFIG}/Cartridges.d/$Store.conf";
        if (-s $ConfFile) {
            tie %CartridgeConf, 'Config::IniFiles', (-file => $ConfFile) or
                Error('FileReadFailed', { 'FileName' => $ConfFile });
        }

        $DatabaseConf{Cartridges} = defined $CartridgeConf{Cartridges}
            ? { %{$CartridgeConf{Cartridges}} } : {};

        # ignore stores without StoreType/Database/datasource
        next unless $DatabaseConf{Connection}{StoreType};
        next unless $DatabaseConf{Connection}{Database};
        next unless $DatabaseConf{Connection}{datasource};

        # ignore store if datasource is not mysql
        next unless $DatabaseConf{Connection}{datasource} =~ /^dbi:mysql/i;
        # ignore store if StoreType is ASPoolDB
        next if $DatabaseConf{Connection}{StoreType} eq 'ASPoolDB';
        # ignore store if StoreType or Database is epagesj
        next if $DatabaseConf{Connection}{StoreType} eq 'epagesj';
        next if $DatabaseConf{Connection}{Database} eq 'epagesj';
        # ignore store if Database is systemdb
        next if $DatabaseConf{Connection}{Database} eq 'systemdb';
        # ignore store if Database is master database
        next if $DatabaseConf{Connection}{Database} eq 'mysql';

        $StoreConf{$Store} = { %DatabaseConf };
    }
    return \%StoreConf;
}

#========================================================================================
# function     DatabaseConfig
# state        public
#----------------------------------------------------------------------------------------
# syntax       $IniConfig = DatabaseConfig($Store);
#----------------------------------------------------------------------------------------
# description  get ini object of the Database.d/$Store.conf
#----------------------------------------------------------------------------------------
# input        $Store | logical database name | string
# return       $IniConfig | config object | object
#========================================================================================
sub DatabaseConfig ($) {
    my ($Store) = @_;
    return GetCachedFile( "$ENV{'EPAGES_CONFIG'}/Database.d/$Store.conf", sub {
        my ($FileName) = @_;
        GetLog->debug( "Loading config file $FileName" );
        return DE_EPAGES::Core::API::IniConfig->new(
            FileName=>$FileName, CaseSensitive => 1, Create => 1 );
    });
}

#========================================================================================
# function     DatabaseConfigFileName
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ConfigFileName= DatabaseConfigFileName($Store);
#----------------------------------------------------------------------------------------
# description  get full file name of the Database.d/$Store.conf
#----------------------------------------------------------------------------------------
# input        $Store | logical database name | string
# return       $ConfigFileName | file name | string
#========================================================================================
sub DatabaseConfigFileName ($) {
    my ($Store) = @_;
    return "$ENV{'EPAGES_CONFIG'}/Database.d/$Store.conf";
}

#========================================================================================
# function     DatabaseConnectionData
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hSection = DatabaseConnectionData($Store, $hFlags);
# example      $hSection = DatabaseConnectionData('Store');
#----------------------------------------------------------------------------------------
# description  Gets database info from Database.d/$Store.conf.
#               Throws an error if the section does not exists.
#               Passwords are decrypted transparently.
#----------------------------------------------------------------------------------------
# input        $Store | logical database name | string
# input        $hFlags | flags that change the behavior of DatabaseConnectionData: <ul>
#                   <li>incomplete - if true, do not throw error if $Store.conf is incomplete
#               </ul> | ref.hash
# return       $hSection | connection info keys, e.g.: <ul>
#                   <li>datasource - complete DBI connection string - string
#                   <li>login - database login name - string
#                   <li>passwd - database password (not encrypted)- string
#               </ul> | ref.hash.string
#========================================================================================
sub DatabaseConnectionData ($;$) {
    my ($Store, $hFlags) = @_;
    TestString('Store', $Store);

    my $IniConfig = DatabaseConfig($Store);

    if (not $IniConfig->existsSection( 'Connection' )) {
        return {} if $hFlags->{'incomplete'};
        Error( 'MissingConfigSection' => {
            Section => 'Connection',
            FileName => DatabaseConfigFileName($Store)
        } );
    }

    my $hSection = $IniConfig->section( 'Connection' );
    return {} if $hFlags->{'incomplete'} and not defined $hSection->{'datasource'};
    return {} if $hFlags->{'incomplete'} and $hSection->{'datasource'} !~ /\S/;

    # Decrypt password if it is encrypted.
    if (defined $hSection->{'encpasswd'}) {
        $hSection->{'passwd'} = dECRYPTdbpASSWORD($hSection->{'encpasswd'});
    }

    my ($Driver, @Params);
    if( $hSection->{'datasource'} =~ /^dbi:
                                      (\w+)  # database driver, e.g. 'mysql'
                                      (?:
                                        :
                                        (.*) # parameters
                                      )?
                                     /x )
    {
        $Driver = $1;
        my $Parameters = $2 // '';

        # Filter Parameters; it should not contain a database name!
        @Params = map {
                /^database=/
                ? do { Warning('DatabaseDefined', { datasource => $hSection->{'datasource'} } ); () }
                : $_
            }
            split (/;/, $Parameters)
    }
    else {
        Error('VoidConfigKey', { FileName => DatabaseConfigFileName($Store),
                                 Section => 'Connection',
                                 Key => 'datasource',
                                 Value => $hSection->{'datasource'} });
    }

    my $Database = $hSection->{'Database'};
    push @Params, "database=$Database" if defined $Database;

    $hSection->{'datasource'} = "dbi:$Driver:" . join(';', @Params);

    Error( 'MissingConfigKey' => {
        Key => 'datasource',
        Section => 'Connection',
        FileName => DatabaseConfigFileName($Store)
    } ) unless defined $hSection->{'datasource'};

    return $hSection;
}

#========================================================================================
# function     GetRootConnections
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $aConnections = GetRootConnections();
#----------------------------------------------------------------------------------------
# description  Gets connections with user=root as defined in $Backup.conf ordered by importance:
#               1. Site.conf exist, find $Backup.conf with Backup=$Backup in Site.conf
#               2. Store.conf exist, find $Backup.conf with Backup=$Backup in Store.conf
#               3. find Backup.conf with login=root and Database=mysql
#               4. find $Backup.conf with login=root and Database=mysql
#----------------------------------------------------------------------------------------
# return       $aConnections | connections as defined in Shared/Config/Database.d/$Backup.conf | ref.array
#========================================================================================
sub GetRootConnections {
    my $aStoreConnections = StoreConnections();

    my @BackupStores;
    RunOnDirectory( "$ENV{EPAGES_CONFIG}/Database.d", sub {
        push @BackupStores, DatabaseConfig('Site')->get( 'Connection', 'Backup' )
            if 'Site' ~~ @$aStoreConnections;
        push @BackupStores, DatabaseConfig('Store')->get( 'Connection', 'Backup' )
            if 'Store' ~~ @$aStoreConnections;
        push @BackupStores, 'Backup' if 'Backup' ~~ @$aStoreConnections;
        push @BackupStores, map { s,\.conf,,; $_ ~~ @$aStoreConnections } glob('Backup*.conf');
        push @BackupStores, @$aStoreConnections;
    });

    my @FoundStores;
    foreach my $Store (@BackupStores) {
        my $IniConfig = DatabaseConfig($Store);
        next unless $IniConfig->get('Connection', 'login');
        next unless $IniConfig->get('Connection', 'login') eq 'root';
        next unless $IniConfig->get('Connection', 'Database');
        next unless $IniConfig->get('Connection', 'Database') eq 'mysql';
        next if $Store ~~ @FoundStores;
        push @FoundStores, $Store;
    }
    return \@FoundStores;
}

#========================================================================================
# function     DatabaseBackupData
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hSection = DatabaseBackupData($Store, $BackupStore);
# example      $hSection = DatabaseBackupData('Store', 'Backup');
#----------------------------------------------------------------------------------------
# description  Gets database backup info from Database.d/$BackupStore.conf.
#               Passwords are decrypted transparently.
#----------------------------------------------------------------------------------------
# input        $Store | logical database name | string
# input        $BackupStore | database backup connection (optional) | string
# return       $hSection | connection info keys, e.g.: <ul>
#                   <li>datasource - dbd connection string - string
#                   <li>login - database login name - string
#                   <li>passwd - database password (not encrypted)- string
#               </ul> | ref.hash.string
#========================================================================================
sub DatabaseBackupData (;$$) {
    my ($Store, $BackupStore) = @_;

    $BackupStore = DatabaseConfig($Store)->get( 'Connection', 'Backup' ) unless $BackupStore;
    return {} unless $BackupStore;

    my $IniConfig = DatabaseConfig($BackupStore);
    return {} unless $IniConfig->existsSection( 'Connection' );

    my $hSection = $IniConfig->section( 'Connection' );

    # Decrypt password if it is encrypted.
    $hSection->{passwd} = dECRYPTdbpASSWORD($hSection->{encpasswd})
        if defined $hSection->{encpasswd};
    $hSection->{StoreName} = $BackupStore;

    return $hSection;
}

#========================================================================================
# function     dECRYPTdbpASSWORD
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Password = dECRYPTdbpASSWORD($EncPassword);
#----------------------------------------------------------------------------------------
# description  decrypt the password to save it at the Database.d/$Store.conf.
#----------------------------------------------------------------------------------------
# input        $EncPassword | crypted database password (ini keyword is encpasswd) | string
# return       $Password | database password (ini keyword is passwd) | string
#========================================================================================
sub dECRYPTdbpASSWORD ($) {
    my ($EncPassword) = @_;

    # hexstring -> binstring
    my $EncodedString = Hex2Str($EncPassword);

    # decrypt binarystring
    my @DES_KEYS = ('Go09rOia', 'u377ureu', 'i1.6nanm', 'd.1Genb!');
    my $Password = des_3ede_cbc_encrypt($EncodedString, @DES_KEYS, 0 );
    $Password = (split /\0/, $Password)[0];
    return $Password;
}

#========================================================================================
# function     DatabaseRedirect
# state        public
#----------------------------------------------------------------------------------------
# syntax       $FileName = DatabaseRedirect($Store);
# syntax       $FileName = DatabaseRedirect($Store, $ConnectionName, $RemoteAddr);
#----------------------------------------------------------------------------------------
# description  Get database redirect file from Database.d/$Store.conf, section $Store,
#               key "redirect", except if $RemoteAddr matches the list of exceptions
#               specified by key "AllowFrom".
#               If $FileName is defined, the content of the file
#               "$ENV{'EPAGES_WEBROOT'}/$FileName"
#               shall be returned the client and no database connection shall be opened.
#----------------------------------------------------------------------------------------
# input        $Store | logical database name | string
# input        $ConnectionName | unused | string
# RemoteAddr   $RemoteAddr | (optional) comma-separated list of IP addresses for which
#                   access is granted anyway | string
# return       $FileName | file name (relative to $EPAGES_WEBROOT), undef if
#                           no redirect is required | string
#========================================================================================
sub DatabaseRedirect ($;$$) {
    my ($Store, undef, $RemoteAddr) = @_;
    TestString('Store', $Store);
    my $IniConfig = DatabaseConfig($Store);

    Error( 'MissingConfigSection' => {
        Section => 'Connection',
        FileName => DatabaseConfigFileName($Store)
    } ) unless $IniConfig->existsSection( 'Connection' );

    my $hSection = $IniConfig->section( 'Connection' );
    my $FileName = $hSection->{'redirect'};

    return undef unless defined $FileName;
    return undef if _IsRemoteAddrAllowed( $RemoteAddr, $hSection->{'AllowFrom'} );
    return $FileName;
}

#========================================================================================
# function     _IsRemoteAddrAllowed
# state        private
#----------------------------------------------------------------------------------------
# syntax       $IsAllowed = _IsRemoteAddrAllowed($RemoteAddr, $AllowFrom);
#----------------------------------------------------------------------------------------
# description  Returns true if $RemoteAddr is included in $AllowFrom.
#----------------------------------------------------------------------------------------
# input        $RemoteAddr | IP address | string
# input        $AllowFrom | allow from list | string array
# return       $RemoteAddr | (optional) client IP address  | string
# return       $AllowFrom | (optional) comma-separated list of allowed IP addresses | string
# return       $IsAllowed | true if allowed | boolean
#========================================================================================
sub _IsRemoteAddrAllowed($$) {
    my ($RemoteAddr, $AllowFrom) = @_;
    LogDebug( '_IsRemoteAddrAllowed', [$RemoteAddr, $AllowFrom] );

    return 0 unless defined $RemoteAddr;
    return 0 unless defined $AllowFrom;

    my @Exceptions = trim split(/,/, $AllowFrom);

    return 1 if $RemoteAddr ~~ @Exceptions;
}

1;
