#==============================================================================
# package      DE_EPAGES::Database::API::DatabaseInstaller
# base         DE_EPAGES::XML::API::XMLInstaller
# state        public
# description  install/patch/uninstall with the database directory, the
#               directory is defined by functions <function databaseDirectory>,
#               EPAGES_CARTRIDGES/<function DE_EPAGES::Core::API::BaseInstaller::cartridgeDirectory>/Database/<function DE_EPAGES::Database::API::DBI::driverName>
#==============================================================================
package DE_EPAGES::Database::API::DatabaseInstaller;
use base DE_EPAGES::XML::API::XMLInstaller;

use strict;
use DE_EPAGES::Core::API::Error qw( Error );
use DE_EPAGES::Core::API::File qw( GetFileContent );
use DE_EPAGES::Core::API::Log qw( GetLog );
use DE_EPAGES::Core::API::Scramble qw( Unscramble );
use DE_EPAGES::Core::API::PerlTools qw( IndexOf );
use DE_EPAGES::Core::API::TestParameter qw( TestString );
use DE_EPAGES::Database::API::Connection qw( GetCurrentDBHandle );
use DE_EPAGES::Database::API::DBInterface qw( ProcessSqlGo );
use DE_EPAGES::Database::API::OrderFile qw( GetSqlFileOrder );

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::Database::API::DatabaseInstaller->new(%Options);
#----------------------------------------------------------------------------------------
# description  Creates a new database installer object.
#----------------------------------------------------------------------------------------
# input        %Options | option
#               <ul>
#                   <li>CartridgeDirectory - cartridge base directory (DE_EPAGES/Database) - string
#                   <li>HasDbSchema - database driver directory will be processed (optional default=1) - boolean
#               </ul> | hash
#========================================================================================
sub new {
    my ($class, %options) = @_;
    $options{'HasDbSchema'} = 1 unless defined $options{'HasDbSchema'};
    unless( $options{'StoreName'} ) {
        my $dbi = GetCurrentDBHandle();
        $options{'StoreName'} = (defined $dbi) ? $dbi->storeName : undef;
    }
    $options{'StoreType'} = $options{'StoreType'};
    my $self = __PACKAGE__->SUPER::new(%options);
    return bless $self, $class;
}

#========================================================================================
# function     install
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->install;
#----------------------------------------------------------------------------------------
# description  Calls <function DE_EPAGES::XML::API::XMLInstaller::install> and
#               installes database object by calling <function installDBStructure>.
#========================================================================================
sub install ($) {
    my $self = shift;

    $self->SUPER::install;
    $self->installDBStructure;
}

#========================================================================================
# function     installDBStructure
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->installDBStructure;
#----------------------------------------------------------------------------------------
# description  Creates database objects, such as tables, views and stored procedures.
#               Calls <function processDatabaseDir> if <function hasDBSchema> returns
#               true and the <function databaseDirectory> exists.
#========================================================================================
sub installDBStructure {
    my $self = shift;

    my $DatabaseDir = $self->databaseDirectory;
    return unless $self->hasDBSchema;
    return unless -d $DatabaseDir;
    GetLog->info(sprintf('install database %s on store %s.', $self->cartridgePackage, $self->storeName));
    $self->processDatabaseDir( $DatabaseDir );
}

#========================================================================================
# function     uninstall
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->uninstall;
#----------------------------------------------------------------------------------------
# description  Calls <function DE_EPAGES::Core::API::FileInstaller::uninstall> and
#               cleanc the database with the database directory of the current driver.
#========================================================================================
sub uninstall ($) {
    my $self = shift;

    my $DatabaseDir = $self->databaseDirectory;
    if( $self->hasDBSchema && -d $DatabaseDir ) {
        GetLog->info(sprintf('uninstall database %s on store %s.', $self->cartridgePackage, $self->storeName));
        $self->cleanupDatabaseDir( $DatabaseDir );
    }

    $self->SUPER::uninstall;
}

#========================================================================================
# function     encryptData
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Installer->encryptData;
#----------------------------------------------------------------------------------------
# description  Encrpyts data of this cartridge.
#========================================================================================
sub encryptData {
    # overload this method in derived cartridges
}

#========================================================================================
# function     hasDBSchema
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hasShema = $Installer->hasDBSchema;
#----------------------------------------------------------------------------------------
# description  Gets information if the cartridge has additional tables or not.
#----------------------------------------------------------------------------------------
# return       $hasShema | cartridge database driver directory will be processed | boolean
#========================================================================================
sub hasDBSchema {
    my $self = shift;
    return $self->{'HasDbSchema'};
}

#========================================================================================
# function     databaseDirectory
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Directory = $Installer->databaseDirectory;
#----------------------------------------------------------------------------------------
# description  Returns sql install/uninstall directory, for example
#               /epages/Cartridges/COMPANY/CARTRIDGE/Database/Sybase.
#----------------------------------------------------------------------------------------
# return       $Directory | directory where the sql files stored | string
#========================================================================================
sub databaseDirectory {
    my $self = shift;
    my $dbi = GetCurrentDBHandle();
    return join '/', $ENV{'EPAGES_CARTRIDGES'}, $self->cartridgeDirectory , 'Database', $dbi->driverName;
}

#========================================================================================
# function     processDatabaseDir
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->processDatabaseDir( $Dir )
# example      $Installer->processDatabaseDir( 'Database/admindb' )
#----------------------------------------------------------------------------------------
# description  Processes SQL files and templates in the specified directory into
#               the current database.
#----------------------------------------------------------------------------------------
# input        $Dir | database directory (e.g. 'Database/admindb') | string
#========================================================================================
sub processDatabaseDir ($) {
    my $self = shift;
    my ($Dir) = @_;
    GetLog()->debug("processDatabaseDir( $Dir )" );
    my $aFiles = GetSqlFileOrder( $Dir );
    $self->processSingleFile( $_ ) foreach @$aFiles;
}

#========================================================================================
# function     processSingleFile
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->processSingleFile( $FileName );
#----------------------------------------------------------------------------------------
# description  Processes a single SQL file into the current database (filename includes /delete_).
#----------------------------------------------------------------------------------------
# input        $FileName | file name | string
#========================================================================================
sub processSingleFile ($) {
    my $self = shift;
    my ($filename) = @_;
    return 0 if $filename =~ /\/delete_/;
    $filename =~ /\.(\w+)$/;
    my $extension = $1;

    if ($extension eq 'sql') {
        $self->processSqlFile( $filename );
    }
    elsif( $extension eq 'xsql' ) {
        $self->processXsqlFile( $filename );
    }
    else {
        GetLog->info("file $filename ignored (no sql,xsql file).");
        return 0;
    }
    return 1;
}

#========================================================================================
# function     cleanupDatabaseDir
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->->cleanupDatabaseDir( $Dir )
# example      $Installer->->cleanupDatabaseDir( 'Database/admindb' )
#----------------------------------------------------------------------------------------
# description  Processes SQL files and templates in the specified directory into
#               the current database.
#----------------------------------------------------------------------------------------
# input        $Dir | database directory (e.g. 'Database/admindb') | string
#========================================================================================
sub cleanupDatabaseDir ($) {
    my $self = shift;
    my ($Dir) = @_;
    GetLog()->debug("cleanupDatabaseDir( $Dir )" );
    my $aFiles = GetSqlFileOrder( $Dir );
    $self->cleanupSingleFile( $_ ) foreach reverse @$aFiles;
}

#========================================================================================
# function     cleanupSingleFile
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->cleanupSingleFile( $FileName );
#----------------------------------------------------------------------------------------
# description  Processes a single SQL file into the current database (filename includes with /delete_).
#----------------------------------------------------------------------------------------
# input        $FileName | file name | string
#========================================================================================
sub cleanupSingleFile ($) {
    my $self = shift;
    my ($filename) = @_;
    return 0 unless $filename =~ /\/delete_/;

    $filename =~ /\.(\w+)$/;
    my $extension = $1;
    if ($extension eq 'sql') {
        $self->processSqlFile( $filename, {'ignore_error'=> $self->force } );
    }
    elsif( $extension eq 'xsql' ) {
        $self->processXsqlFile( $filename, {'ignore_error'=> $self->force } );
    }
    else {
        GetLog->info("file $filename ignored (no sql,xsql file).");
        return 0;
    }

    return 1;
}
#========================================================================================
# function     processXsqlFile
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->processXsqlFile( $FileName, $hFlags );
#----------------------------------------------------------------------------------------
# description  Processes a .xsql files into the current database. The file content
#               is unscrambled before execution.
#----------------------------------------------------------------------------------------
# input        $FileName | xsql file name | string
# input        $hFlags | processing flag can be set 'ignore_error' (optional)| ref.hash
#========================================================================================
sub processXsqlFile ($;$) {
    my $self = shift;
    my ($FileName, $hFlags) = @_;
    GetLog()->debug("processXsqlFile($FileName)");

    my $rContent = GetFileContent( $FileName );
    my $Content = Unscramble( $$rContent );

    ProcessSqlGo( $Content, $hFlags );
}

#========================================================================================
# function     processSqlFile
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->processSqlFile( $FileName, $hFlags );
#----------------------------------------------------------------------------------------
# description  Processes a .sql files into the current database.
#----------------------------------------------------------------------------------------
# input        $FileName | sql file name | string
# input        $hFlags | processing flag can be set 'ignore_error' (optional)| ref.hash
#========================================================================================
sub processSqlFile ($;$) {
    my $self = shift;
    my ($FileName, $hFlags) = @_;
    GetLog()->debug("processSqlFile($FileName)");

    my $rContent = GetFileContent( $FileName );
    ProcessSqlGo( $$rContent, $hFlags );
}

#========================================================================================
# function     onPatchProcessSingleFile
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->onPatchProcessSingleFile($File, $hParams);
#----------------------------------------------------------------------------------------
# example      $Installer->onPatchProcessSingleFile(
#                         'Database/Sybase/Procedures/i_Cartridge.sql',
#                         {'CVS' => 'Changed',
#                         });
#----------------------------------------------------------------------------------------
# description  Called by PatchImportHandler by Method onPatchProcessSingleFile.
#               Process sql at file $File if file includes $dbi->driverName.
#----------------------------------------------------------------------------------------
# input        $File | cartridgefilename without cartridge directory | string
# input        $hParams | patch file parameters, keys are
#               <ul>
#                   <li>CVS          - could be Deleted, New or Changed - string
#                   <li>Force        - process is forced (not stops) (optional default 0)- boolean
#               </ul> | ref.hash.*
#========================================================================================
sub onPatchProcessSingleFile {
    my $self = shift;
    my ( $File, $hParams ) = @_;
    my $dbiDriverName= GetCurrentDBHandle()->driverName;
    if ( $File !~ /\/$dbiDriverName|SQL\//) {
        GetLog->info("file $File ignored (does not  belong to current driver: " . $dbiDriverName .  '.');
    }else {
        my $Force = exists $hParams->{'Force'} ? $hParams->{'Force'} : $self->force;
        my $FileName = join '/', $ENV{'EPAGES_CARTRIDGES'}, $self->cartridgeDirectory, $File;
        $FileName =~ /\.(\w+)$/;
        my $extension = $1;
        if ($extension eq 'sql') {
            $self->processSqlFile( $FileName, {'ignore_error'=> $Force } );
        }
        elsif( $extension eq 'xsql' ) {
            $self->processXsqlFile( $FileName, {'ignore_error'=> $Force} );
        }
        else {
            GetLog->info("file $FileName ignored (no sql,xsql file).");
        }
    }
    return;
}

#========================================================================================
# function     startPatchDBStructure
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->startPatchDBStructure;
#----------------------------------------------------------------------------------------
# description  Writes a log entry when the database structure update starts.
#========================================================================================
sub startPatchDBStructure {
    my $self = shift;

    GetLog->info(sprintf('patch db structure cartridge %s on store %s.', $self->cartridgePackage, $self->storeName));
    return;
}

#========================================================================================
# function     finishPatchDBStructure
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->finishPatchDBStructure;
#----------------------------------------------------------------------------------------
# description  Writes a log entry when the database structure is up to date.
#========================================================================================
sub finishPatchDBStructure {
    my $self = shift;

    GetLog->info(sprintf('db structure cartridge %s is patched to version %s.', $self->cartridgePackage, $self->version));
    return;
}

#========================================================================================
# function     doPatchDBStructure
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->doPatchDBStructure;
# syntax       $Installer->doPatchDBStructure('6.11.1');
#----------------------------------------------------------------------------------------
# description  Patches cartridge db structure in the active database.
#               Calls methods named "patchDBStructure_$DBStructureVersion", where
#               $DBStructureVersion is the actual db structure version of the cartridge
#               ('.' are replaced with '_') until the current cartridge version is
#               reached.
#               Calls <function DE_EPAGES::XML::API::XMLInstaller::importPatchDir>("Patch_$DBStructureVersion")
#               if the function patchDBStructure_$DBStructureVersion does not exist.
#----------------------------------------------------------------------------------------
# input        $Version | goal version number, current number is chosen if undefined | string
#========================================================================================
sub doPatchDBStructure {
    my $self = shift;
    my ($Version) = @_;

    #if no version is given, the version from Installer is chosen
    $Version = $self->version unless defined $Version;

    GetLog->info(sprintf('patch db structure cartridge %s on store %s.', $self->cartridgePackage, $self->storeName));

    my $DBVersion = $self->getDBStructureVersion;

    #get available patches
    my $aPatchVersions = $self->patches;

    #check if patch version is already current version
    unless ($Version eq $self->version()){
        #check if given version exists in cartridge versions
        if (not (grep {$_ eq $Version} @$aPatchVersions)) {
            GetLog->info(
                'warning cartridge version not found while doPatchDBStructure Package => ' .
                $self->cartridgePackage .
                ' OldVersion=> ' .
                $DBVersion .
                ' NewVersion=> ' .
                $Version .
                "\npatch db structure to current version => " .
                $self->version
            );

            #set patch version to current version
            $Version = $self->version;
        }
    }

    while ($DBVersion ne $Version) {
        my $i = IndexOf($DBVersion, @$aPatchVersions);
        Error('NO_DBSTRUCTUREPATCH_FOUND', { 'Package'=>$self->cartridgePackage, 'OldVersion'=>$DBVersion, 'NewVersion'=>$Version})
            if $i == -1; #DBVersion is not included at available patches

        my $PatchSubDir = "Patch_$DBVersion";
        $PatchSubDir =~ s/\./_/g;

        my $FunctionName = "patchDBStructure_$DBVersion";
        $FunctionName =~ s/\./_/g;
        if ($self->can($FunctionName) ) {
            $DBVersion = $self->$FunctionName;
        }
        else {
            $DBVersion = $self->importPatchDir( $PatchSubDir, 'PatchDBStructure' );
            unless( defined $DBVersion ) {
                $DBVersion = (scalar(@$aPatchVersions) <= $i+1) ? $Version : $aPatchVersions->[$i+1];
            }
        }

        $self->setDBStructureVersion($DBVersion);
    }
    return;
}

#========================================================================================
# function     getDBStructureVersion
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DBStructureVersion = $Installer->getDBStructureVersion;
#----------------------------------------------------------------------------------------
# description  Returns the current database structure version.
#               Actually returns $self->installedVersion.
#               This method is overloaded by <package DE_EPAGES::Cartridge::API::CartridgeInstaller>.
#========================================================================================
sub getDBStructureVersion {
    my $self = shift;

    return $self->installedVersion;
}

#========================================================================================
# function     setDBStructureVersion
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->setDBStructureVersion( $Version );
#----------------------------------------------------------------------------------------
# description  Sets the current database structure version.
#               Actually does nothing.
#               This method is overloaded by <package DE_EPAGES::Cartridge::API::CartridgeInstaller>.
#========================================================================================
sub setDBStructureVersion {
    my $self = shift;
    my ($Version) = @_;
    TestString('Version', $Version);

    # do nothing
    return;
}

#========================================================================================
# function     patchDBStructure
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Installer->patchDBStructure;
# syntax       $Installer->PatchDBStructure('6.11.1');
#----------------------------------------------------------------------------------------
# description  Patches the database structure of the cartridge in the current dataabase
#               to the newest level.
#               The default implementation calls <function startPatchDBStructure>,
#               <function doPatchDBStructure> and <function finishPatchDBStructure>
#               if the cartridge is installed.
#----------------------------------------------------------------------------------------
# input        $Version | goal version number, current number is chosen if undefined | string
#========================================================================================
sub patchDBStructure {
    my $self = shift;
    my ($Version) = @_;

    if ($self->isInstalled) {
        $self->startPatchDBStructure();
        $self->checkAntiquePatches();
        $self->doPatchDBStructure($Version);
        $self->finishPatchDBStructure();
    } else {
        GetLog->info(sprintf('No patchDBStructure %s on store %s, because cartridge is not installed.', $self->cartridgePackage, $self->storeName));
    }
}

#========================================================================================
# function     getCurrentDBHandle
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $dbi = $Installer->getCurrentDBHandle();
#----------------------------------------------------------------------------------------
# description  Returns the current database handle.
#----------------------------------------------------------------------------------------
# return       $dbi | database handle object | DE_EPAGES::Database::API::DBI
#========================================================================================
sub getCurrentDBHandle{
    my $self = shift;

    return GetCurrentDBHandle();
}

1;
