use strict;
use Config;
use Date::Format qw(time2str);
use File::Basename;
use Filesys::DfPortable qw(dfportable);
use Getopt::Long;
use Readonly;
use Sys::Hostname;
use DE_EPAGES::Core::API::Constants qw(MEBI);
use DE_EPAGES::Core::API::File qw( CreateDirectory );
use DE_EPAGES::Core::API::PerlTools qw( ElementOf );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Core::API::SystemCommand qw( SystemCommand );
use DE_EPAGES::Database::API::BackupMySQL qw( GetStoreConf MySQLShowVariables GetDatabaseSize );
use DE_EPAGES::Database::API::Config qw(StoreConnections DatabaseConnectionData DatabaseBackupData);

# possible values for option -target
my @TargetList = qw(restore_db initial_backup_db final_backup_db full_backup_db);

sub Main {
    local $| = 1;

    my %Options = ();
    my @GetOpt = qw(help fullbackup logbackup restore storename=s connection=s nooutput target=s args=s all own compress:-1);

    usage() unless GetOptions(\%Options, @GetOpt);
    usage() if defined $Options{help};

    my ($Errors, $Dump, $BackupType, $InitialBackupDate);
    my $Target    = defined $Options{target} ? basename($Options{target}) : '';
    my $TargetDir = defined $Options{target} ? dirname($Options{target})  : '';

    if (defined $Options{target}) {
        usage("Invalid target: $Options{target}/$Target") unless ElementOf($Target, @TargetList);
        if ( -f "$Options{target}" ) {
            print "INFO: $Target skipped, because file exists: $Options{target}\n";
            exit 0;
        }
        $Dump = $Target eq 'restore_db' ? 'Restore' : 'Backup';

        if ($Dump eq 'Restore') {
            $InitialBackupDate = InitialBackupFound($TargetDir, \%Options) or exit 0;
        }
    }
    elsif( $Options{restore} ) {
        $Dump = 'Restore';
    }
    elsif( $Options{fullbackup} ) {
        $Dump = 'Backup';
        $BackupType = '-full';
    }
    elsif( $Options{logbackup} ) {
        $Dump = 'Backup';
        $BackupType = '-log';
    }
    else {
        usage( 'One of the parameters is missing: -target|fullbackup|logbackup|restore' );
    }

    my $StoreList = $Options{storename} ? [$Options{storename}] : StoreConnections();

    HandleDump($_, $BackupType, $Target, $Dump, $InitialBackupDate, \%Options, \$Errors) foreach @$StoreList;

    if ($Errors) {
        print STDERR "\nEXITING: Database $Dump Failed:\n";
        print STDERR "======================================\n";
        die $Errors;
    }

    UnlinkTargets($TargetDir, $Dump, \%Options) if defined $Options{target} and $Target ne 'full_backup_db';
}

sub InitialBackupFound {
    my ($TargetDir, $Options) = @_;

    my $InitialBackupFile = -f "$TargetDir/initial_backup_db" ? "$TargetDir/initial_backup_db" : '';
    if ($Options->{own} and not $InitialBackupFile) {
        $InitialBackupFile = dirname($TargetDir). '/initial_backup_db' if -f dirname($TargetDir). '/initial_backup_db';
    }
    if ( not $InitialBackupFile or not -f $InitialBackupFile ) {
        print "INFO: no DB restore: $TargetDir/initial_backup_db: no such file\n";
        return;
    }
    open my $File, '<', $InitialBackupFile or die "Error reading file '$InitialBackupFile'";
    my $InitialBackupDate = (<$File>)[0];
    close $File or die "Error closing file '$InitialBackupFile'";
    chomp($InitialBackupDate);

    return $InitialBackupDate;
}

sub UnlinkTargets {
    my ($TargetDir, $Dump, $Options) = @_;

    chdir $TargetDir or die "$TargetDir: $!";

    if ($Dump eq 'Backup') {
        unlink('restore_db');
        if ($Options->{all}) {
            unlink ("$_/restore_db") foreach <*>;
        }
    } else {
        unlink('initial_backup_db');
        unlink('final_backup_db');
    }

    my $TimeStamp = time2str('%Y%m%d%H%M%S', time());

    open my $File, '>>', basename($Options->{target}) or
        die "Cannot write $Options->{target}:$!";
    print $File $TimeStamp;
    close $File or die "Cannot close file $File:$!";
}

sub HandleDump {
    my ($Store, $BackupType, $Target, $Dump, $InitialBackupDate, $Options, $Errors) = @_;

    my $hBackupSection = DatabaseBackupData($Store, $Options->{connection});
    my $BackupConnection = $hBackupSection->{StoreName};

    return unless $BackupConnection;

    my $hSection = DatabaseConnectionData($Store);

    return if defined $hSection->{active} and not $hSection->{active};
    return unless $hSection->{datasource} =~ /^dbi:mysql/;
    return unless $hSection->{Database};

    my $LocalDatabase = GetStoreConf($Store)->{LocalDB};

    # go on if option own is defined but database isn't local
    return if not $LocalDatabase and $Options->{own};
    $LocalDatabase = 1 if $Options->{all};

    my $BackupHost = $hSection->{BackupHost};
    $BackupHost = $hBackupSection->{BackupHost} unless $BackupHost;
    $BackupHost = 'localhost' unless $BackupHost;

    # go on if host isn't localhost or myself
    return unless $BackupHost eq hostname() or $BackupHost =~ /^(ipv6-)?(localhost|127.0.0.1|::1)$/;

    # go on if host is localhost and database isn't local
    return if $BackupHost =~ /^(ipv6-)?(localhost|127.0.0.1|::1)$/ and not $LocalDatabase;

    my $ScriptArgs = $Options->{connection} ? "-connection $Options->{connection}" : '';
    $ScriptArgs .= " -storename $Store";
    $ScriptArgs .= ' -nooutput' if $Options->{nooutput};

    my $DumpArgs = $Options->{args};
    $DumpArgs = $hSection->{"${Dump}MySQLArgs"} unless $DumpArgs;
    $DumpArgs = $hBackupSection->{"${Dump}MySQLArgs"} unless $DumpArgs;
    $ScriptArgs .= " $DumpArgs" if $DumpArgs;

    # check BackupType
    $BackupType = $BackupType ? " $BackupType" : CheckBackupType($Store, $hSection, $Target, $Dump, $Errors);

    if( $Dump eq 'Backup' ) {
        return unless $BackupType;
        $ScriptArgs .= " $BackupType";
        $ScriptArgs .= " -compress $Options->{compress}" if defined $Options->{compress};
    } elsif ($Dump eq 'Restore') {
        $ScriptArgs .= ' -nobinlogs' if $BackupType !~ /-log/;
        $ScriptArgs .= " -dumpdate $InitialBackupDate" if $InitialBackupDate;
    }

    # check free disk space
    if( $Dump eq 'Backup' ) {
        my $DumpDir = GetStoreConf($Store)->{DumpDir};
        $DumpDir = -d "$ENV{EPAGES}/MySQLDumps" ? "$ENV{EPAGES}/MySQLDumps" : $ENV{EPAGES}
        unless $DumpDir;
        CreateDirectory($DumpDir);

        Readonly::Scalar my $SpareFactor => 1.2;
        my $RequiredSpace = GetDatabaseSize($Store) * $SpareFactor;
        my $FreeSpace = dfportable($DumpDir)->{bavail};
        if ($RequiredSpace > $FreeSpace) {
            my $RequiredSpaceMB = int($RequiredSpace / MEBI);
            my $FreeSpaceMB = int($FreeSpace / MEBI);
            $$Errors .= "\nERROR: $Store: Not enough disk space in dump directory to dump database.\n".
                "  Dump directory: $DumpDir\n  Required space: $RequiredSpaceMB MB\n  Free space: $FreeSpaceMB MB\n";
            return;
        }
    }

    my $Script = "$ENV{EPAGES_CARTRIDGES}/DE_EPAGES/Database/Scripts/MySQL$Dump.pl";
    $Script = "Include/DE_EPAGES/Database/Scripts/MySQL$Dump.pl"
        if -f "Include/DE_EPAGES/Database/Scripts/MySQL$Dump.pl";

    my $PerlExe = $Config{'osname'} eq 'MSWin32' ? $^X : $ENV{'PERL'};
    my $Command = FindMySQLDumpInPath($Store). "\"$PerlExe\" -IInclude $Script $ScriptArgs";
    SystemCommand( $Command, { 'Print' => !$Options->{nooutput} } );
}

sub FindMySQLDumpInPath {
    my ($Store) = @_;

    if ($Config{'osname'} eq 'MSWin32') {
        foreach my $Dir (split /;/, $ENV{PATH}) {
            return '' if -f "$Dir/mysqldump.exe";
        }
    } else {
        foreach my $Dir (split /:/, $ENV{PATH}) {
            return '' if -x "$Dir/mysqldump";
        }
    }

    my $BaseDir = MySQLShowVariables($Store)->{basedir};
    $BaseDir =~ s,[\\/]+,/,g;
    $BaseDir =~ s,[\\/]+$,,;

    if ($Config{'osname'} eq 'MSWin32' && -f "$BaseDir/bin/mysqldump.exe") {
        $BaseDir =~ s,/,\\,g;
        return "set PATH=%PATH%;$BaseDir\\bin& ";
    } elsif (-x "$BaseDir/bin/mysqldump") {
        return "PATH=\$PATH:$BaseDir/bin ";
    }
    return '';
}

sub CheckBackupType {
    my ($Store, $hSection, $Target, $Dump, $Errors) = @_;

    return ' -full' if $Target eq 'full_backup_db';

    my $MomentType = $Target eq 'initial_backup_db' ? 'Initial' : 'Final';
    $MomentType .= 'BackupType';

    if ($hSection->{BackupType} !~ /\S/ and $hSection->{$MomentType} !~ /\S/ ) {
        print "INFO: no \L$Dump\E: $Store: keys $MomentType and BackupType: empty\n";
        return;
    }
    if ($hSection->{$MomentType} =~ /\S/) {
        if ($hSection->{$MomentType} !~ /^-(?:log|full)$/) {
            $$Errors .= "\nERROR: $Store: $hSection->{$MomentType}: void $MomentType, use -full or -log\n";
            return;
        }
        return " $hSection->{$MomentType}";
    }
    if ($hSection->{BackupType} !~ /^-(?:log|full)$/) {
        $$Errors .= "\nERROR: $Store: $hSection->{BackupType}: void BackupType, use -full or -log\n";
        return;
    }
    return " $hSection->{BackupType}";
}

sub usage {
    my ($Message) = @_;

    print "$Message\n\n" if $Message;

    my $TargetsString = join ', ', @TargetList;

    print <<END_USAGE;
Usage:
    perl $0 {-fullbackup|logbackup|restore | -target path/to/TARGET} OPTIONS

Actions:
    -fullbackup         backup the database
    -logbackup          backup the binlogs
    -restore            restore the database
    -target             TARGET must be one of: $TargetsString

Options:
    -args               arguments for backup/restore script to execute
                        (overwrites Backup/RestoreMySQLArgs in Database.d/*.conf)
    -storename          store name for backup    (default: all local databases)
    -connection         store name for the database connection       (optional)
    -all                dump/restore even if no running MySQL found on
                        localhost                                    (optional)
    -own                dump/restore only if own connection          (optional)
    -help               show the command line options                (optional)
    -nooutput           do not print any output                      (optional)

Example:
    perl $0 -fullbackup -args "-prefix foobar"
    perl $0 -restore
    perl $0 -target fullbackup
END_USAGE
    exit 2;
}

RunScript(
    'Sub' => \&Main
);
