#========================================================================================
# package      DE_EPAGES::Order::API::Object::ContainerShippingMethod
# state        public
# base         DE_EPAGES::Shipping::API::Object::ShippingMethod
# description  Object interface for ContainerShippingMethod extends the shipping method
#               with line item functionalities.
#========================================================================================
package DE_EPAGES::Order::API::Object::ContainerShippingMethod;
use base qw( DE_EPAGES::Shipping::API::Object::ShippingMethod );

use strict;
use DE_EPAGES::Customer::API::Shop qw (GetCustomerForUser);

#========================================================================================
# function     lineItem
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $hInfo = $ShippingMethod->lineItem($Container);
#----------------------------------------------------------------------------------------
# description  Returns information for adding a new shipping line item.
#               Called for new baskets if this shipping methos is default or for
#               existing baskets when the shipping method is changed.
#----------------------------------------------------------------------------------------
# input        $Container | line item container | object
# return       $hInfo | shipping line item info, keys
#               <ul>
#                   <li>LineItemPrice - price for shipping is 0 - float
#                   <li>Name -  name of shipping method
#                   <li>Quantity - quantity is 1 - float
#                   <li>ShippingMethod - self - object
#                   <li>TaxClass - tax class of shipping - object
#                   <li>TaxRate - tax rate - float
#               </ul> | ref.hash
#========================================================================================
sub lineItem {
    my $self = shift;
    my ($Container) = @_;

    my $TaxClass = $self->get('TaxClass');
    my $TaxRate = defined $TaxClass ? $Container->taxRate($TaxClass) : 0;

    return {
        'LineItemPrice'     => $Container->formatter->round($self->calculateShippingPrice($Container)),
        'Name'              => $self->get('NameOrAlias', $Container->get('LanguageID')),
        'Quantity'          => 1,
        'ShippingMethod'    => $self,
        'TaxClass'          => $TaxClass,
        'TaxRate'           => $TaxRate,
    };
}

#========================================================================================
# function     canAddToBasket
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Can = $ShippingMethod->canAddToBasket($hVars);
#----------------------------------------------------------------------------------------
# description  Tests if a shipping method can be added to the basket with the given currency.
#               Uses function <function DE_EPAGES::Shipping::API::Object::ShippingMethod::price>
#               to evaluate if shipping methode has a price (price can be 0). Possible for
#               methods, which costs, but have no price in the given currency.
#----------------------------------------------------------------------------------------
# input        $hVars | vars of container and basket | ref.hash
# return       $Can | can added to the basket | boolean
#========================================================================================
sub canAddToBasket {
    my $self = shift;
    my ($hVars) = @_;

    return 0 unless $self->get('IsActivated');
    return 0 unless $self->get('ShippingType')->get('IsActivated');

    #test regions
    my $Region= $self->get('Region');
    if (defined $Region) {
        $hVars->{'ShippingCountryID'} = $hVars->{'LineItemContainer'}->get('ShippingCountryID') if defined $hVars->{'LineItemContainer'} and not defined $hVars->{'ShippingCountryID'};
        if (defined $hVars->{'ShippingCountryID'}) {
            return 0 unless grep {$_->{'CountryID'} == $hVars->{'ShippingCountryID'}} @{$Region->get('Countries')};
        }
    }

    # test customer groups
    my $CustomerGroups = $self->get('CustomerGroups');
    if (scalar @$CustomerGroups) {
        my $CustomerGroup = $hVars->{'CustomerGroup'};
        if(not defined $CustomerGroup and defined $hVars->{'LineItemContainer'}) {
            my $ContainerParent = $hVars->{'LineItemContainer'}->businessContainer->parent;
            my $Customer = $ContainerParent->get('Customer');
            if (not defined $Customer) {
                my $User = $ContainerParent->get('User');
                $Customer = GetCustomerForUser($User, $ContainerParent->getSite()) if defined $User;
            }
            $CustomerGroup = $Customer->get('CustomerGroup') if defined $Customer;
        }
        $CustomerGroup = $CustomerGroups->[0]->getSite()->get('NonMemberCustomerGroup') unless defined $CustomerGroup;
        $hVars->{'CustomerGroup'} = $CustomerGroup;
        return 0 if not defined $CustomerGroup or not grep { $_->id == $CustomerGroup->id } @$CustomerGroups;
    }

    return $self->usableAtStorefront($hVars);
}

#========================================================================================
# function     addToContainer
#----------------------------------------------------------------------------------------
# syntax       $LineItem = $ShippingMethod->addToContainer($Container);
#----------------------------------------------------------------------------------------
# description  Adds shipping and tax line item to container.
#----------------------------------------------------------------------------------------
# input        $Container | line item container | object
# return       $LineItem | line item | object
#========================================================================================
sub addToContainer {
    my $self = shift;
    my ($Container) = @_;
    return $Container->insertLineItem($self->get('ShippingType')->get('LineItemClass'), $self->lineItem($Container));
}

#========================================================================================
# function     recalculateLineItem
#----------------------------------------------------------------------------------------
# syntax       $ShippingMethod->recalculateLineItem($LineItem);
#----------------------------------------------------------------------------------------
# description  Recalculates the price of an existing shipping line item.
#----------------------------------------------------------------------------------------
# input        $LineItem | line item | object
# input        $hValues | changed values (optional) | ref.hash
#========================================================================================
sub recalculateLineItem {
    my $self = shift;
    my ($LineItem, $hValues) = @_;
    my $Container = $LineItem->container;
    $hValues = {} unless defined $hValues;

    my $ahDeliveryLineItems = $hValues->{'DeliveryLineItem'};

    if ( !(defined $ahDeliveryLineItems && exists $ahDeliveryLineItems->[0]) ) {
        if ( !($hValues->{'BasketPatch'}) ||
            (defined $hValues->{'BasketPatch'}->{'RecShipping'} && $hValues->{'BasketPatch'}->{'RecShipping'}) )
        {
            my $NewLineItemPrice = $Container->formatter->round($self->calculateShippingPrice($Container));
            $LineItem->set({'LineItemPrice'=>$NewLineItemPrice}) if $NewLineItemPrice != $LineItem->get('LineItemPrice');
        }
    }
    if (defined $hValues->{'BasketPatch'}) {
        delete($hValues->{'BasketPatch'});
    }
    return;
}

#========================================================================================
# function     calculateShippingPrice
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $ShippingPrice = $ShippingMethod->calculateShippingPrice($Container);
#----------------------------------------------------------------------------------------
# description  Returns the shipping price for the given line item container. This
#               method is used by <function lineItem> and <function recalculateLineItem>
#               unless they are overloaded. The default implementation returns 0.
#               Override this method to change the shipping price calculation of a
#               basket or order.
#               To calculate the shipping costs without a lineitem container, e.g. for
#               export to product portals, also overload the method
#               <function DE_EPAGES::Shipping::API::Object::ShippingMethod::price>.
#----------------------------------------------------------------------------------------
# input        $Container | LineItemContainer | ref.hash
# return       $ShippingPrice | price of shipping | money
#========================================================================================
sub calculateShippingPrice {
    return 0;
}

#========================================================================================
# function     valueToFreeDelivery
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $MissingValueForFreeDelivery = $ShippingMethod->valueToFreeDelivery($Container);
#----------------------------------------------------------------------------------------
# description  Returns the additional basket values required to get free delivery.
#               Override this method if you want to display a message in the basket
#               indicating the additional value required to get free shipping.
#               If the return value is greater that 0, such a message is displayed in
#               the basket page.
#----------------------------------------------------------------------------------------
# input        $Container | LineItemContainer | ref.hash
# return       $MissingValueForFreeDelivery| missing value | money
#========================================================================================
sub valueToFreeDelivery {
    return 0;
}

1;
