#========================================================================================
# package      DE_EPAGES::ShopMigration::API::ShortURL
# state        public
#----------------------------------------------------------------------------------------
# description  inserts redirect entries for contentpages/legalpages/products
#               so former shorturls still work post migration
#========================================================================================
package DE_EPAGES::ShopMigration::API::ShortURL;
use base qw(Exporter);

use strict;
use DE_EPAGES::Content::API::Table::ShortUrlRedirection qw(GetInfoShortUrlRedirectionsByContentObjectID);
use DE_EPAGES::Core::API::Error qw(Error);
use DE_EPAGES::Core::API::Log qw(LogDebug);
use DE_EPAGES::Core::API::PerlTools qw(GenerateGUID);
use DE_EPAGES::Database::API::Config qw(DatabaseConnectionData);
use DE_EPAGES::Database::API::Connection qw(GetCurrentDBHandle RunOnStore);
use DE_EPAGES::Object::API::Domain qw(GetUnitySubDomain);
use DE_EPAGES::Object::API::Factory qw(LoadRootObject);
use DE_EPAGES::Object::API::Language qw ( GetCodeByLanguageID );
use DE_EPAGES::Object::API::Table::Locale qw(GetLanguageByLocaleID);
use DE_EPAGES::ShopMigration::API::Constants qw($FILTER_SHORTURL);
our @EXPORT_OK = qw(
    InsertRedirect
    ExistsRedirect
    IsAllowedShortURL
    GetAllUILocalesForShop
    AddDeletedShortUrlRedirections
);

#========================================================================================
# function     InsertRedirect
# state        private
#----------------------------------------------------------------------------------------
# syntax       InsertRedirect($Shop, $FromSlug, $ToSlug)
#----------------------------------------------------------------------------------------
# description  inserts a redirect for in epages-ui, does nothing if redirect already
#               exists or there is no shop for the given locale
#               Due to a bug we have to insert the redirect for all locales/shops
#               known to epages-ui
#----------------------------------------------------------------------------------------
# input        $Shop         | baseshop or prepshop | object
# input        $FromSlug     | the shorturl/URI in epages6 | string
# input        $ToSlug       | the target slug in epages-ui | string
#========================================================================================
sub InsertRedirect {
    my ($Shop, $FromSlug, $ToSlug) = @_;

    return unless (defined $FromSlug && defined $ToSlug);
    unless (IsAllowedShortURL($FromSlug)) {
        LogDebug(sprintf("Skipped invalid shorturl redirect '%s' for shop '%s'\n", $Shop->alias, $FromSlug), 1);
        return;
    }
    my $Store = LoadRootObject()->get('StoreName');
    my $ShopSlug = GetUnitySubDomain($Shop->alias);
    my $UIStore = DatabaseConnectionData($Store)->{InstanceUnityUI};
    if (!defined $UIStore || !$UIStore) {
        Error('UI_STORE_NOT_DEFINED', { Store => $Store })
    } else {
        RunOnStore(
            Store => $UIStore,
            Sub   => sub {
                my $dbi = GetCurrentDBHandle();
                # shop exisiting for locale?
                my $aaShopIDs = $dbi->execute(<<SQL,
SELECT id FROM shops
WHERE shopGateID = ( SELECT id FROM shopGates WHERE slug=? );
SQL
                    [ $ShopSlug ]);
                unless (scalar @$aaShopIDs) {
                    LogDebug(sprintf("No shop found for shopgate (slug) '%s'\n",
                        $ShopSlug,), 1);
                    return;
                }
                foreach my $Row (@$aaShopIDs) {
                    my $ShopID = $Row->[0];
                    # Redirect already existing ?
                    my $aaRows = $dbi->execute(<<SQL,
SELECT COUNT(1) FROM redirects WHERE fromSlug = ? and type = 'shorturl' and shopId = ?
SQL
                        [ $FromSlug, $ShopID ],
                        [ 'Integer' ]);
                    if ($aaRows->[0][0] > 0) {
                        LogDebug(sprintf("Found existing redirect entry for shopid '%s' fromslug '%s' in epages-ui\n",
                            $ShopID, $FromSlug), 1);
                        return;
                    } else {
                        my $RedirectID = GenerateGUID();
                        $dbi->do(<<SQL,
INSERT INTO redirects (id, fromSlug, toSlug, type, createdAt, updatedAt, shopId)
VALUES (?, ?, ?, 'shorturl', NOW(), NOW(), ?);
SQL
                            [ $RedirectID, $FromSlug, $ToSlug, $ShopID ],
                        );
                        LogDebug(sprintf("Inserted redirect entry in epages-ui for shop '%s' ('%s' -> '%s')\n",
                            $ShopID, $FromSlug, $ToSlug), 1);
                    }
                }
            }
        );
    }
}

#========================================================================================
# function     GetAllUILocalesForShop
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $aLocales = GetAllUILocalesForShop($Shop)
#----------------------------------------------------------------------------------------
# description  returns an array of locales for which a shop exists in epages-ui
#----------------------------------------------------------------------------------------
# input        $Shop         | baseshop or prepshop | object
# return       $aLocales     | list of locales | ref array
#========================================================================================
sub GetAllUILocalesForShop {
    my ($Shop) = @_;

    my @Locales = ();
    my $Store = LoadRootObject()->get('StoreName');
    my $ShopSlug = GetUnitySubDomain($Shop->alias);
    my $UIStore = DatabaseConnectionData($Store)->{InstanceUnityUI};
    if (!defined $UIStore || !$UIStore) {
        Error('UI_STORE_NOT_DEFINED', { Store => $Store })
    } else {
        RunOnStore(
            Store => $UIStore,
            Sub   => sub {
                my $dbi = GetCurrentDBHandle();
                my $aaRows = $dbi->execute(<<SQL,
SELECT locale from shops
where shopgateid IN (SELECT id FROM shopGates WHERE slug=?);
SQL
                    [ $ShopSlug ]);
                foreach my $Row (@$aaRows) {
                    push @Locales, $Row->[0];
                }
            }
        );
    }
    return \@Locales;
}

#========================================================================================
# function     ExistsRedirect
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $Exists = ExistsRedirect($Shop, $FromSlug)
#----------------------------------------------------------------------------------------
# description  Checks if the given redirect already exists in epages-ui for any
#               locale / shopid
#----------------------------------------------------------------------------------------
# input        $Shop         | baseshop or prepshop | object
# input        $FromSlug     | the shorturl/URI in epages6 | string
# return       $Exists       | redirect exists ? | Boolean
#========================================================================================
sub ExistsRedirect {
    my ($Shop, $FromSlug) = @_;

    my $Exists = 0;
    my $Store = LoadRootObject()->get('StoreName');
    my $ShopSlug = GetUnitySubDomain($Shop->alias);
    my $UIStore = DatabaseConnectionData($Store)->{InstanceUnityUI};
    if (!defined $UIStore || !$UIStore) {
        Error('UI_STORE_NOT_DEFINED', { Store => $Store })
    } else {
        RunOnStore(
            Store => $UIStore,
            Sub   => sub {
                my $dbi = GetCurrentDBHandle();
                my $aaShopID = $dbi->execute(<<SQL,
SELECT id FROM shops
WHERE shopGateID = ( SELECT id FROM shopGates WHERE slug=? );
SQL
                    [ $ShopSlug ]);
                return unless scalar @$aaShopID;
                foreach my $Row (@$aaShopID) {
                    my $ShopID = $Row->[0];
                    my $aaRows = $dbi->execute(<<SQL,
SELECT COUNT(1) FROM redirects
WHERE fromSlug = ? and type = 'shorturl' and shopId = ?
SQL
                        [ $FromSlug, $ShopID ],
                        [ 'Integer' ]);
                    $Exists = $aaRows->[0][0] > 0 ? 1 : 0;
                    last if $Exists;
                }
            }
        );
    }
    return $Exists;
}

#========================================================================================
# function     IsAllowedShortURL
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $Allowed = IsAllowedShortURL($FromSlug)
#----------------------------------------------------------------------------------------
# description  Checks if the given shorturl is possible in epages-ui due
#               to reserved paths for product, category and legal pages
#----------------------------------------------------------------------------------------
# input        $FromSlug     | the shorturl/URI in epages6 | string
# return       $Allowed      | shorturl allowed ? | Boolean
#========================================================================================
sub IsAllowedShortURL {
    my ($FromSlug) = @_;

    return ($FromSlug =~ $FILTER_SHORTURL) ? 0 : 1;
}

#========================================================================================
# function     AddDeletedShortUrlRedirections
# state        public
#----------------------------------------------------------------------------------------
# syntax       AddDeletedShortUrlRedirections($Shop);
#----------------------------------------------------------------------------------------
# description  adds the entries from shorturlredirection (shorturls from deleted objects)
#               we cannot do this per category because they might collide with new/
#               existing shorturls from non-deleted objects so we must finish import first
#----------------------------------------------------------------------------------------
# input        $Shop         | baseshop or prepshop | object
#========================================================================================
sub AddDeletedShortUrlRedirections {
    my ($Shop) = @_;

    LogDebug(sprintf("AddDeletedShortUrlRedirections for %s\n", $Shop->alias), 1);
    my $SubCategories = $Shop->get('Categories')->get('RecursiveSubCategories');
    unshift @$SubCategories, $Shop->get('Categories');
    my $StartPageID = $Shop->get('Categories')->id;
    my $aLocales = GetAllUILocalesForShop($Shop);
    foreach my $Category (@$SubCategories) {
        # also add deleted shorturls to epages-ui (localized startpage instead of 404)
        foreach my $hShortUrlRedirection (@{GetInfoShortUrlRedirectionsByContentObjectID($Category->id)}) {
            foreach my $LocaleID (@$aLocales) {
                my $LanguageID = GetLanguageByLocaleID($LocaleID);
                next unless ($hShortUrlRedirection->{'LanguageID'} == $LanguageID);
                my $FromSlug = $hShortUrlRedirection->{Uri};
                if (defined $FromSlug) {
                    # UNITY-7056: point to StartPage if referenced contentobject is StartPage otherwise use guid
                    my $ToSlug = '/' . GetCodeByLanguageID($LanguageID) .
                        (($hShortUrlRedirection->{ContentObjectID} == $StartPageID) ? '' : '/c/' . $Category->guid);
                    InsertRedirect($Shop, $FromSlug, $ToSlug);
                }
            }
        }
    }
}

1;
